package jp.co.sds_corp.http.server;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import javax.servlet.http.HttpServletResponse;

import jp.co.sds_corp.util.config.IConfigurable;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

public class Document implements Context, IConfigurable
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	String path;
	String url;
	List<String> listIdx = new ArrayList<String>();
	Map<String,String> mapHeader = new HashMap<String,String>();
	Map<String,MimeType> mapMime = new HashMap<String,MimeType>();
	Map<String,Content> cache = new ConcurrentHashMap<String,Content>();

	protected Document() {
	}

	@Override
	public void init(IConfiguration config) throws Exception {
		String var;
		var = config.getAttributes().get("path");
		if (!var.endsWith("/")) {
			var = var + "/";
		}
		path = var;
		var = config.getAttributes().get("url");
		if (!var.endsWith("/")) {
			var = var + "/";
		}
		url = var;

		initIndex(config);
		initHeader(config);
		IConfiguration mimeCfg = config.getChild("mimeType");
		initMime(mimeCfg);
	}

	void initIndex(IConfiguration config) throws Exception {
		List<IConfiguration> list = config.listChild("index");
		for (IConfiguration idxCfg : list) {
			String var = idxCfg.getAttributes().get("file", null);
			if (var != null && var.length() > 0) {
				listIdx.add(var);
			}
		}
	}

	void initHeader(IConfiguration config) throws Exception {
		List<IConfiguration> list = config.listChild("header");
		for (IConfiguration hdCfg : list) {
			String name = hdCfg.getAttributes().get("name", null);
			String value = hdCfg.getAttributes().get("value", null);
			if (name == null || name.length() == 0 ||
				value == null || value.length() == 0)
			{
				continue;
			}
			mapHeader.put(name, value);
		}
	}

	void initMime(IConfiguration config) throws Exception {
		String file = config.getAttributes().get("config", null);
		if (file != null && file.length() > 0) {
			try {
				readMime(file);
			} catch (Exception ex) {
				throw(new Exception("mime config read error.", ex));
			}
		}

		List<IConfiguration> list = config.listChild("mime");
		for (IConfiguration hdCfg : list) {
			String type = hdCfg.getAttributes().get("type");
			String ext = hdCfg.getAttributes().get("extention");
			MimeType mime = new MimeType(type, ext);
			mapMime.put(ext, mime);
			logger.trace(this, "add mime " + type + "," + ext);
		}
	}

	void readMime(String config) throws Exception {
		BufferedReader rd = new BufferedReader(new FileReader(config));
		try {
			int no = 0;
			for ( ;; ) {
				String line = rd.readLine();
				if (line == null) {
					break;
				}
				no++;
				int idx = line.indexOf("#");
				if (idx >= 0) {
					line = line.substring(0, idx);
				}
				line = line.trim();
				if (line.length() == 0) {
					continue;
				}
				String[] arr = line.split("[ \t]+");
				if (arr.length < 2) {
					logger.warn(this, "illegal format line(" + no + ") " + line);
					continue;
				}
				String type = arr[0];
				for (int i = 1; i < arr.length; i++) {
					String ext = arr[i];
					MimeType mime = new MimeType(type, ext);
					mapMime.put(ext, mime);
					logger.trace(this, "add mime " + type + "," + ext);
				}
			}
		} finally {
			try {
				rd.close();
			} catch (Exception ex) {
			}
		}
	}

	@Override
	public boolean match(HttpRequest request) {
		String query = request.getQueryString();
		if (query != null && query.startsWith(getURL())) {
			return true;
		}
		return false;
	}

	Content getContent(HttpRequest request) throws Exception {
		String query = request.getQueryString();
		if (!match(request)) {
			throw(new FileNotFoundException(query + " not found."));
		}
		Content content = cache.get(query);
		if (content != null) {
			return content;
		}

		if (!query.startsWith("/")) {
			throw(new IllegalArgumentException("illegal query " + query));
		}
		String real = query.replace(getURL(), "");
		real = path + real;

		File file = null;
		if (real.endsWith("/")) {
			for (String idx : listIdx) {
				file = new File(real + idx);
				if (file.exists()) {
					break;
				}
			}
		} else {
			file = new File(real);
		}
		content = Content.getContent(query, file);
		MimeType mime = findMime(content);
		if (mime != null) {
			content.setMimeType(mime);
			logger.debug(this, query + " -> " + file + " " + mime.getType());
		} else {
			logger.trace(this, query + " -> " + file + " NONE");
		}
		cache.put(query, content);
		return content;
	}

	@Override
	public void service(HttpRequest request, HttpResponse response)
		throws Exception
	{
		response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
		response.putHeaders(mapHeader);
		response.setProtocol(request.getProtocol());

		Content content = null;
		try {
			content = getContent(request);
			response.setContent(content);
			if (content != null) {
				MimeType mime = content.getMimeType();
				if (mime != null) {
					response.putHeader("Content-Type", mime.getType());
				}
			}
			response.setStatus(HttpServletResponse.SC_OK);
		} catch (FileNotFoundException ex) {
			logger.error(this, "file not found", ex);
			response.setStatus(HttpServletResponse.SC_NOT_FOUND);
		} catch (IllegalArgumentException ex) {
			logger.error(this, "illegal argument", ex);
			response.setStatus(HttpServletResponse.SC_FORBIDDEN);
		} catch (Exception ex) {
			logger.error(this, "error", ex);
			response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
		}
	}

	MimeType findMime(Content content) {
		String url = content.getFile().getName();
		int idx = url.lastIndexOf(".");
		if (idx < 0) {
			return null;
		}
		String ext = url.substring(idx + 1);
		MimeType mime = mapMime.get(ext);
		return mime;
	}

	@Override
	public void start() {
	}

	@Override
	public void stop() {
		cache.clear();
	}

	void setPath(String var) {
		this.path = var;
	}
	public String getPath() {
		return this.path;
	}

	void setURL(String var) {
		this.url = var;
	}
	String getURL() {
		return this.url;
	}

	public String printCache() {
		StringBuilder buf = new StringBuilder();
		buf.append(toString()).append("\n");
		buf.append("cache={\n");
		for (Map.Entry<String,Content> entry : cache.entrySet()) {
			buf.append("\t").append(entry.getValue()).append("\n");
		}
		buf.append("}");
		return buf.toString();
	}

	public String printMime() {
		StringBuilder buf = new StringBuilder();
		buf.append(toString()).append("\n");
		buf.append("mime={\n");
		for (Map.Entry<String,MimeType> entry : mapMime.entrySet()) {
			buf.append("\t").append(entry.getKey()).append("=")
				.append(entry.getValue()).append("\n");
		}
		buf.append("}");
		return buf.toString();
	}

	public String printConfig() {
		StringBuilder buf = new StringBuilder();
		buf.append(toString()).append("\n");
		buf.append("header={\n");
		for (Map.Entry<String,String> entry : mapHeader.entrySet()) {
			buf.append("\t").append(entry.getKey()).append("=")
				.append(entry.getValue()).append("\n");
		}
		buf.append("}");

		return buf.toString();
	}

	@Override
	public String toString() {
		StringBuilder buf = new StringBuilder();
		buf.append(getClass().getName()).append(" url=").append(url)
			.append(" path=").append(path);
		buf.append(" index=[");
		for (String index : listIdx) {
			buf.append(index).append(",");
		}
		buf.append("]");
		buf.append(" cache size=").append(cache.size());

		return buf.toString();
	}

}
