package jp.co.sds_corp.http.server;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletResponse;

import jp.co.sds_corp.util.config.ConfigUtil;
import jp.co.sds_corp.util.config.DOMConfiguration;
import jp.co.sds_corp.util.config.IConfigurable;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

public class HttpServer implements IConfigurable
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	IConfiguration doc;
	String address;
	int port;
	ContextLoader contextLoader;
	List<Context> listContext;
	Map<SocketChannel,RequestParser> mapReq = new HashMap<SocketChannel,RequestParser>();
	Server server;
	Thread srvThr;
	boolean start = false;
	boolean stop = false;
	Selector selector;
	static final long TIMEOUT = 30000;

	public void init() throws Exception {
		doc = DOMConfiguration.findResource( "httpserver.xml" );
		init( doc );
	}

	@Override
	public void init( IConfiguration config ) throws Exception {
		address = config.getAttributes().get("address", null);
		port = ConfigUtil.getIntAttribute(config, "port");

		contextLoader = new ContextLoader();
		contextLoader.init(config);
		listContext = contextLoader.listContext();
	}

	public void start() throws Exception {
		logger.info(this, "start");
		setStart(false);
		setStop(false);
		for (Context context : listContext) {
			context.start();
		}
		server = new Server();
		srvThr = new Thread(server);
		srvThr.setName("HttpServerThread");
		srvThr.start();
		synchronized(this) {
			if (isStart()) {
				return;
			}
			wait(10000);
		}
		if (!isStart()) {
			logger.error(this, "not started.");
		}
	}

	public void stop() throws Exception {
		logger.info(this, "stop");
		setStop(true);
		try {
			srvThr.join();
		} catch (Exception ex) {
		}
		for (Context context : listContext) {
			context.stop();
		}
	}

	class Server implements Runnable {
		public void run() {
			ServerSocketChannel srvCh;
			try {
				srvCh = startServer();
			} catch (Exception ex) {
				logger.error(this, "start error", ex);
				return;
			} finally {
				setStart(true);
			}
			try {
				select();
			} catch (Exception ex) {
				logger.error(this, "select error", ex);
			}
			try {
				srvCh.close();
			} catch (Exception ex) {
				logger.error(this, "close error", ex);
			}
			logger.info(this, "stopped.");
		}
	}

	ServerSocketChannel startServer() throws Exception {
		logger.info(this, "startServer");
		selector = Selector.open();
		ServerSocketChannel srvCh = ServerSocketChannel.open();
		srvCh.configureBlocking(false);

		logger.info(this, "bind " + address + ":" + port);
		if (address != null && address.length() > 0) {
			InetAddress addr = InetAddress.getByName(address);
			srvCh.socket().bind(new InetSocketAddress(addr, port));
		} else {
			srvCh.socket().bind(new InetSocketAddress(port));
		}
		srvCh.register(selector, SelectionKey.OP_ACCEPT);
		return srvCh;
	}

	void select() throws Exception {
		for ( ;; ) {
			if (isStop()) {
				break;
			}
			int ret = selector.select(1000);
			Iterator<SelectionKey> itr = selector.selectedKeys().iterator();
			while (itr.hasNext()) {
				SelectionKey key = itr.next();
				itr.remove();
				if (key.isAcceptable()) {
					accept((ServerSocketChannel )key.channel());
				} else
				if (key.isReadable()) {
					try {
						read((SocketChannel )key.channel());
					} catch (Exception ex) {
						logger.error(this, "read error", ex);
					}
				}
			}
			try {
				checkTimeout();
			} catch (Exception ex) {
				logger.error(this, "error", ex);
			}
		}
	}

	void checkTimeout() throws Exception {
		Date now = new Date();
		Iterator<SocketChannel> itr = mapReq.keySet().iterator();
		while (itr.hasNext()) {
			SocketChannel channel = itr.next();
			RequestParser parser = mapReq.get(channel);
			if (parser == null) {
				itr.remove();
				continue;
			}
			Date last = parser.getLastReceived();
			if (last == null || (now.getTime() - last.getTime()) > TIMEOUT) {
				logger.warn(this, "timeout " + parser.getRemoteAddr());
				parser.getChannel().close();
				itr.remove();
			}
		}
	}

	void accept(ServerSocketChannel srvCh) {
		try {
			SocketChannel channel = srvCh.accept();
			if (channel == null) {
				logger.warn(this, "channel is null");
				return;
			}
			String addr = "N/A";
			SocketAddress sockAddr = channel.socket().getRemoteSocketAddress();
			if (sockAddr != null) {
				addr = sockAddr.toString();
			}
			logger.trace(this, "connect from " + addr);
			channel.configureBlocking(false);
			channel.register(selector, SelectionKey.OP_READ);
		} catch (Exception ex) {
			logger.error(this, "accept error", ex);
		}
	}

	void closeChannel(SocketChannel channel) throws Exception {
		channel.close();
		mapReq.remove(channel);
	}

	void read(SocketChannel channel) throws Exception {
		RequestParser parser = mapReq.get(channel);
		if (parser == null) {
			parser = new RequestParser(channel);
			mapReq.put(channel, parser);
		}
		String addr = parser.getRemoteAddr();
		HttpRequest request = null;
		try {
			int ret = parser.read();
			request = parser.parse();
			if (request == null) {
				if (ret < 0) {
					logger.trace(this, "unexpected EOF:" + addr + "\n"
						+ parser);
					closeChannel(channel);
					throw(new Exception("unexpected EOF"));
				}
				return;
			}
		} catch (Exception ex) {
			logger.error(this, "read error:" + addr, ex);
			closeChannel(channel);
			return;
		}

		logger.trace(this, "request:" + addr + " "
			+ request.getMethod() + " " + request.getQueryString() + " "
			+ request.getProtocol());
/*
		logger.trace("request:" + addr + "\n" + request);
*/
		HttpResponse response = getResponse(request);

		Date end = new Date();
		logger.detail(this, addr + " "
			+ request.getMethod() + " "
			+ request.getQueryString() + " "
			+ request.getProtocol() + " "
			+ response.getStatusCode() + " "
			+ response.getHeader("Content-Type") + " "
			+ response.getContentLength() + " "
			+ (end.getTime() - parser.getStartAt().getTime()) );
		ByteBuffer res = ByteBuffer.wrap(response.getBytes());
/*
		logger.trace("response:" + addr + "\n" + response);
		Charset charset = Charset.defaultCharset();
		logger.trace("response:" + addr + "\n" + charset.decode(res));
		res.flip();
*/
		channel.write(res);
		closeChannel(channel);
	}

	HttpResponse getResponse(HttpRequest request) {
		HttpResponse response = new HttpResponse();
		response.setStatus(HttpServletResponse.SC_NOT_FOUND);

		try {
			Context context = contextLoader.findContext(request);
			if (context != null) {
				context.service(request, response);
			}
		} catch (Exception ex) {
			response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
		}
		response.setProtocol(request.getProtocol());
		return response;
	}

	public synchronized void setStart(boolean var) {
		this.start = var;
		notifyAll();
	}
	public synchronized boolean isStart() {
		return this.start;
	}

	public synchronized void setStop(boolean var) {
		this.stop = var;
	}
	public synchronized boolean isStop() {
		return this.stop;
	}

	public String printCache() {
		StringBuilder buf = new StringBuilder();
		for (Context context : listContext) {
			if (context instanceof Document) {
				Document doc = (Document )context;
				buf.append(doc.getClass().getName()).append(":");
				buf.append(doc.printCache()).append("\n");
			}
		}
		return buf.toString();
	}

	public String printMime() {
		StringBuilder buf = new StringBuilder();
		for (Context context : listContext) {
			if (context instanceof Document) {
				Document doc = (Document )context;
				buf.append(doc.getClass().getName()).append(":");
				buf.append(doc.printMime()).append("\n");
			}
		}
		return buf.toString();
	}

	public String printConfig() {
		StringBuilder buf = new StringBuilder();
		buf.append("Listen:").append(address).append(".")
			.append(port).append("\n");
		for (Context context : listContext) {
			if (context instanceof Document) {
				Document doc = (Document )context;
				buf.append(doc.getClass().getName()).append(":");
				buf.append(doc.printConfig()).append("\n");
			}
		}
		return buf.toString();
	}

	public static void main(String[] args) {
		try {
			HttpServer server = new HttpServer();
			server.init();
			server.start();
		} catch (Exception ex) {
			logger.error(HttpServer.class, "HttpServer error", ex);
		}
	}

}
